
#include <src/Core/ImageStream.h>
#include "FFmpegRecorder.h"

namespace dyncam {

FFmpegStreamRecorder::~FFmpegStreamRecorder() {
    if (m_frameNumber > 0)
        stop();
}

bool FFmpegStreamRecorder::start(std::string filename,
                                 cv::Size frameSize,
                                 int openCVType,
                                 ushort maxFPS,
                                 EventsT<cv::Mat> input,
                                 int compress,
                                 bool hardwareAcc) {
    // initialize ffmpeg
    int ret = 0;

#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(58, 9, 100)
    avcodec_register_all();
#endif

    /* find the mpeg1 video encoder */
    if (hardwareAcc) {
        m_codec = avcodec_find_encoder_by_name("nvenc_h264");
        if (m_codec) {
            std::cout << "using hardware acceleration" << std::endl;
        }
    }
    if (!hardwareAcc || !m_codec) {
        m_codec = avcodec_find_encoder_by_name("libx264");
        if (m_codec) {
            std::cout << "using software encoder" << std::endl;
            // unset hardware acceleration explicitly as we could have tried it but it failed
            hardwareAcc = false;
        }
    }

    if (!m_codec) {
        fprintf(stderr, "Codec not found\n");
        return false;
    }

#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(58, 9, 100)
    avcodec_register( m_codec );
#endif

    m_codecContext = avcodec_alloc_context3(m_codec);
    if (!m_codecContext) {
        fprintf(stderr, "Could not allocate video codec context\n");
        return false;
    }

    auto bytePerPixel = static_cast<const unsigned int>(cv::Mat(1, 1, openCVType).elemSize());
    /* put sample parameters */
    //m_codecContext->bit_rate = 1600000;
    /* resolution must be a multiple of two */
    m_codecContext->width = frameSize.width * bytePerPixel / 2;
    m_codecContext->height = frameSize.height;
    /* frames per second */
    m_codecContext->time_base.num = 1;
    m_codecContext->time_base.den = maxFPS;
    /* emit one intra frame every ten frames
    * check frame pict_type before passing frame
    * to encoder, if frame->pict_type is AV_PICTURE_TYPE_I
    * then gop_size is ignored and the output of encoder
    * will always be I frame irrespective to gop_size
    */
    m_codecContext->gop_size = 10;
    m_codecContext->max_b_frames = 5;
    m_codecContext->pix_fmt = AV_PIX_FMT_YUV422P; // 16bpp // AV_PIX_FMT_RGBA;

    if (compress == 0 || hardwareAcc)
        ret = av_opt_set(m_codecContext->priv_data, "preset", "fast", 0);
    else
        ret = av_opt_set(m_codecContext->priv_data, "preset", "ultrafast", 0);
    if (ret < 0) {
        fprintf(stderr, "Could not set preset\n");
        return false;
    }

    ret = av_opt_set(m_codecContext->priv_data, "crf", std::to_string(compress).c_str(), 0);
    if (ret < 0) {
        fprintf(stderr, "Could not set compression\n");
        return false;
    }

    /* open it */
    ret = avcodec_open2(m_codecContext, m_codec, nullptr);
    if (ret < 0) {
        fprintf(stderr, "Could not open codec\n");

        std::cout << "Supported pixel formats:" << std::endl;
        for (unsigned int i = 0; m_codec->pix_fmts[i] != -1; ++i) {
            std::cout << av_get_pix_fmt_name(m_codec->pix_fmts[i]) << std::endl;
        }

        return false;
    }

    m_frame = av_frame_alloc();
    if (!m_frame) {
        fprintf(stderr, "Could not allocate video frame\n");
        return false;
    }
    m_frame->format = m_codecContext->pix_fmt;
    m_frame->width = m_codecContext->width;
    m_frame->height = m_codecContext->height;

    /* the image can be allocated by any means and av_image_alloc() is
    * just the most convenient way if av_malloc() is to be used */
    ret = av_image_alloc(m_frame->data, m_frame->linesize, m_codecContext->width,
                         m_codecContext->height, m_codecContext->pix_fmt, 32);
    if (ret < 0) {
        fprintf(stderr, "Could not allocate raw picture buffer\n");
        return false;
    }

    m_fileOutput.open(filename, std::ios_base::binary);
    if (!m_fileOutput.is_open()) {
        fprintf(stderr, "Unable to open \n");
        return false;
    }

    m_frameNumber = 0;

    m_streamObserver = Observe(input, [this, bytePerPixel, maxFPS](cv::Mat in) {
        // prepare for encoding
        av_init_packet(&m_packet);
        m_packet.data = nullptr;    // packet data will be allocated by the encoder
        m_packet.size = 0;

        // copy the data from the mat to frame
        memcpy(m_frame->data[0], in.data, m_frame->width * m_frame->height * bytePerPixel / 2);

        // calculate the presentation time stamp
        if (m_frameNumber == 0)
            m_startedRecording = std::chrono::high_resolution_clock::now();
        auto now = std::chrono::high_resolution_clock::now();
        auto secsFromStart = std::chrono::duration<double>(now - m_startedRecording);
        m_frame->pts = static_cast<int64_t>(secsFromStart.count() * maxFPS);

        // encode the image
        auto err = avcodec_send_frame(m_codecContext, m_frame);
        if (err < 0) {
            char buf[100];
            std::cout << "AV error message: " << av_make_error_string(buf, 100, err) << std::endl;
            fprintf(stderr, "Error encoding frame\n");
            return;
        }

        err = avcodec_receive_packet(m_codecContext, &m_packet);
        if (err >= 0) {
            m_fileOutput.write(reinterpret_cast<char *>(m_packet.data), m_packet.size);
        } else {
            char buf[100];
            std::cout << "AV error message: " << av_make_error_string(buf, 100, err) << std::endl;
        }

        ++m_frameNumber;
    });

    return true;
}

void FFmpegStreamRecorder::stop() {
    m_streamObserver.Detach();

    // get and write the delayed packages
    av_init_packet(&m_packet);
    m_packet.data = nullptr;    // packet data will be allocated by the encoder
    m_packet.size = 0;

    int i = m_frameNumber;
    int ret = 1;
    for (; ret >= 0; i++) {
        ret = avcodec_receive_packet(m_codecContext, &m_packet);
        if (ret >= 0) {
            m_fileOutput.write(reinterpret_cast<char *>(m_packet.data), m_packet.size);
        }
    }

    uint8_t endcode[] = {0, 0, 1, 0xb7};
    m_fileOutput.write(reinterpret_cast<char *>(endcode), sizeof(endcode));

    avcodec_close(m_codecContext);
    av_free(m_codecContext);
    av_freep(&m_frame->data[0]);
    av_frame_free(&m_frame);

    m_frameNumber = 0;

    m_fileOutput.close();
}

bool FFmpegCamRecorder::start(const std::string &filePrefix, StreamCollection &streams,
                              ushort maxFPS, int compress, bool hardwareAcc) {
    std::ofstream metaDataFile(filePrefix + "_metadata.txt");

	// the intrinsics are in the ir stream
	/*const auto rawIr = streams.GetStreamFirst<ImageStream>(StreamTypes::IR);
	if (rawIr) { 
		auto intrinsicsParentId = rawIr->GetId(); 
		auto intrinsics = streams.GetStreamFirst<IntrinsicsStream>(StreamTypes::INTRINSICS, intrinsicsParentId);
		if (intrinsics)
		{
			std::stringstream os;
			cereal::JSONOutputArchive archive_out(os);
			archive_out(intrinsics);
			metaDataFile << os.str() << std::endl;
		}
	}*/

    for (const auto &stream : streams.GetAllStreams()) {
        auto imgStream = std::dynamic_pointer_cast<ImageStream>(stream);
        if (!imgStream) {
            // the stream is not an imagestream and therefore not handled by the ffmpeg writer
            continue;
		}

        m_streamRecorders.emplace_back();
        auto &recorder = m_streamRecorders.back();
        if (!recorder.start(filePrefix + "_" + imgStream->ToString() + ".h264",
                            imgStream->m_resolution.Value(),
                            imgStream->m_cvPixelType.Value(),
                            maxFPS,
                            imgStream->events,
                            compress,
                            hardwareAcc)) {
            stop();
            return false;
        }

		/*std::stringstream os;
		cereal::JSONOutputArchive archive_out(os);
		archive_out(imgStream);
		metaDataFile << os.str() << std::endl;*/

//        // todo: serialize stream properties
//        metaDataFile << (unsigned int) stream << " " << imgStream.m_resolution.Value().width << " "
//                     << imgStream.m_resolution.Value().height << " "
//                     << imgStream.m_cvPixelType.Value() << " ";
//
//        auto intrinsics = imgStream.m_intrinsics.Value();
//        metaDataFile << intrinsics.cx << " " << intrinsics.cy << " " << intrinsics.fx << " "
//                     << intrinsics.fy << " "
//                     << intrinsics.k1 << " " << intrinsics.k2 << " " << intrinsics.k3 << " "
//                     << intrinsics.p1 << " " << intrinsics.p2 << "\n";

    }
	std::stringstream os;
	cereal::JSONOutputArchive archive_out(os);
	archive_out(streams);
	metaDataFile << os.str() << std::endl;

    metaDataFile.close();

    return true;
}

void FFmpegCamRecorder::stop() {
    for (auto &recorder : m_streamRecorders) {
        recorder.stop();
    }
}

} // namespace dyncam
