//
// Created by maany on 6/26/17.
//

#include "Freenect2Cam.h"
#include "src/Core/ImageStream.h"
#include <libfreenect2/registration.h>
#include <condition_variable>
#include <chrono>

#define DEBUG_IDENTIFIER "<Freenect2Cam.h> : "

using namespace react;

namespace dyncam {


/**
 * Helper to create a libfreenect Frame from a cv Mat. Both share the same memory.
 * @param m Matrix with allocated memory.
 * @return Frame who's data is stored in m
 */
libfreenect2::Frame frameFromMat(const cv::Mat &m) {
    return libfreenect2::Frame(static_cast<size_t>(m.cols),
                        static_cast<size_t>(m.rows),
                        m.elemSize(), m.data);
}

/**
 * Helper to create a cv Mat from a libfreenect Frame. Both share the same memory.
 * We need to pass frame as pointer. Otherwise there is a double free.
 * @param f Frame from libfreenect
 * @param cvType type identifier for the cv Mat
 * @return Matrix who's data is stored in f.
 */
cv::Mat matFromFrame(const libfreenect2::Frame *f, int cvType) {
    return cv::Mat(static_cast<int>(f->height),
                   static_cast<int>(f->width),
                   cvType, f->data);
}

Freenect2Cam::Freenect2Cam(std::string serial, libfreenect2::PacketPipeline *pipeline):
    pipeline_(pipeline),
    serial_(serial) {

    // Needs to be called before getDefaultSerialNumber()
    freenect2_.enumerateDevices();
#if defined(_DEBUG) && defined (_WIN32)
	serial_ = "002742663447"; // Set manual for win10 debug mode. Alter number according to specific camera!
#else
	serial_ = freenect2_.getDefaultDeviceSerialNumber(); 
#endif // _DEBUG
	Streams.SetSerial(serial_);

    // Unique id for this camera. For parent child relation in collection
    int cameraId = Stream::GenerateUniqueId();

    // Add this camera's streams to the collection
    auto colorSize = cv::Size(1920, 1080);
    auto irSize = cv::Size(512, 424);
    colorStream_ = std::make_shared<ImageStream>(StreamTypes::COLOR, cameraId, colorSize, CV_8UC4 );
    depthStream_ = std::make_shared<ImageStream>(StreamTypes::DEPTH, cameraId, irSize,    CV_32FC1);
    irStream_    = std::make_shared<ImageStream>(StreamTypes::IR,    cameraId, irSize,    CV_32FC1);
    colorStream_->events = colorSource_;
    depthStream_->events = depthSource_;
    irStream_->events    = ir_Source;
    Streams.AddStream(colorStream_);
    Streams.AddStream(depthStream_);
    Streams.AddStream(irStream_);

    // intrinsics
    irIntrinsicsStream_    = std::make_shared<IntrinsicsStream>(irStream_->GetId());
    colorIntrinsicsStream_ = std::make_shared<IntrinsicsStream>(colorStream_->GetId());
    Streams.AddStream(irIntrinsicsStream_);
    Streams.AddStream(colorIntrinsicsStream_);

    // processed streams
    undistortedDepthStream_ =
        std::make_shared<ImageStream>(StreamTypes::UNDISTORTED_DEPTH, depthStream_->GetId(), irSize, CV_32FC1);
    registeredColorStream_ =
        std::make_shared<ImageStream>(StreamTypes::REGISTERED_COLOR,  colorStream_->GetId(), irSize, CV_8UC4);
    undistortedDepthStream_->events = undistorteddepthSource_;
    registeredColorStream_->events = registeredColorSource_;
    Streams.AddStream(undistortedDepthStream_);
    Streams.AddStream(registeredColorStream_);

    // produces a pair of registered maps: color and undistorted depth
    // we produce both at the same time because that's how libfreenect works
    registeredMaps_ = Transform(depthStream_->events,
                               [&](const cv::Mat depth) {

                                   std::lock_guard<std::mutex> lock_cv(m_newColorAvailableMutex);
								
                                   // result matrices
                                   auto width = undistortedDepthStream_->m_resolution().width;
                                   auto height = undistortedDepthStream_->m_resolution().height;
                                   cv::Mat registeredColorMat(height, width, CV_8UC4);
                                   cv::Mat undistortedDepthMat(height, width, CV_32FC1);

                                   // libfreenect accesses them through these frame objects
                                   auto registeredFrame = frameFromMat(registeredColorMat);
                                   auto undistortedFrame = frameFromMat(undistortedDepthMat);

                                   // input data
                                   libfreenect2::Frame colorFrame = frameFromMat(m_lastColor);
                                   libfreenect2::Frame depthFrame = frameFromMat(depth);
                                   registration->apply(&colorFrame,
                                                       &depthFrame,
                                                       &undistortedFrame,
                                                       &registeredFrame);

                                   return std::pair<cv::Mat, cv::Mat>(std::move(registeredColorMat),
                                                                      std::move(undistortedDepthMat));
                               });
    D::SetNodeName(*registeredMaps_.ptr_, "registeredMaps");

    // extract single streams for further processing
    registeredColorStream_->events =
        Transform(registeredMaps_, [this](const std::pair<cv::Mat, cv::Mat> &images) {

		return images.first;
        });
    undistortedDepthStream_->events =
        Transform(registeredMaps_, [](const std::pair<cv::Mat, cv::Mat> &images) {
            return images.second;
        });

    D::SetNodeName(*m_state.ptr_, "CameraState");
}

bool Freenect2Cam::Open() {

    // open if not already connected
    if (state != State::DEVICE_OPEN && Connect()) {

        device->setColorFrameListener(this);
        device->setIrAndDepthFrameListener(this);

		libfreenect2::Freenect2Device::IrCameraParams irParams = device->getIrCameraParams();
		libfreenect2::Freenect2Device::ColorCameraParams colorParams = device->getColorCameraParams();

		/*irParams.cx = 256.684;
		irParams.cy = 207.085;
		irParams.fx = 366.193;
		irParams.fy = 366.193;
		irParams.k1 = 0.0893804;
		irParams.k2 = -0.272566;
		irParams.k3 = 0.0958438;
		colorParams.cx = 959.5;
		colorParams.cy = 539.5;
		colorParams.fx = 1081.37;
		colorParams.fy = 1081.37;*/
		colorParams.mx_x0y0 = -0.037082;
		colorParams.mx_x0y1 = 0.00252753;
		colorParams.mx_x0y2 = -1.23812e-005;
		colorParams.mx_x0y3 = 1.84079e-005;
		colorParams.mx_x1y0 = 0.633835;
		colorParams.mx_x1y1 = 0.000352006;
		colorParams.mx_x1y2 = 0.000456852;
		colorParams.mx_x2y0 = -8.48728e-005;
		colorParams.mx_x2y1 = 4.2612e-005;
		colorParams.mx_x3y0 = 0.00549695;
		/*colorParams.my_x0y0 = 0.0587319;
		colorParams.my_x0y1 = 0.633596;
		colorParams.my_x0y2 = 0.000207549;
		colorParams.my_x0y3 = 0.000866004;
		colorParams.my_x1y0 = -0.00224316;
		colorParams.my_x1y1 = 2.99789e-005;
		colorParams.my_x1y2 = 3.23921e-005;
		colorParams.my_x2y0 = -0.000114432;
		colorParams.my_x2y1 = 0.000498586;
		colorParams.my_x3y0 = -1.15959e-005;*/

        registration = std::make_unique<libfreenect2::Registration>(irParams, colorParams);
    }

    //cv::namedWindow("Color2", CV_WINDOW_NORMAL);
    return state == State::DEVICE_OPEN;
}

void Freenect2Cam::Close() {
    Disconnect();
}

std::set<StreamTypes> Freenect2Cam::getStreamCapabilities() {
    return {StreamTypes::COLOR, StreamTypes::IR, StreamTypes::DEPTH,
            StreamTypes::REGISTERED_COLOR, StreamTypes::UNDISTORTED_DEPTH};
}

bool Freenect2Cam::Connect() {
    if (pipeline_ != nullptr) {
        device = freenect2_.openDevice(serial_, pipeline_);
    } else {
        device = freenect2_.openDevice(serial_);
    }

    if (device == nullptr || !device->start()) {
        std::cout << DEBUG_IDENTIFIER << "failure opening device!" << std::endl;
        state = State::DEVICE_FAILED;
    } else {
        state = State::DEVICE_OPEN;

        // intrinsics
        auto intrIR = device->getIrCameraParams();
        CamIntrinsics irIntrinsics(intrIR.fx, intrIR.fy, intrIR.cx, intrIR.cy,
                                   intrIR.k1, intrIR.k2, intrIR.k3, intrIR.p1, intrIR.p2);
        irIntrinsicsStream_->Intrinsics.Set(irIntrinsics);

        auto intrColor = device->getColorCameraParams();
		CamIntrinsics colorIntrinsics(intrColor.fx, intrColor.fy, intrColor.cx, intrColor.cy,
			intrColor.shift_d, intrColor.shift_m,
			intrColor.mx_x3y0, intrColor.mx_x0y3, intrColor.mx_x2y1, intrColor.mx_x1y2, intrColor.mx_x2y0,
			intrColor.mx_x0y2, intrColor.mx_x1y1, intrColor.mx_x1y0, intrColor.mx_x0y1, intrColor.mx_x0y0,
			intrColor.my_x3y0, intrColor.my_x0y3, intrColor.my_x2y1, intrColor.my_x1y2, intrColor.my_x2y0,
			intrColor.my_x0y2, intrColor.my_x1y1, intrColor.my_x1y0, intrColor.my_x0y1, intrColor.my_x0y0);
        colorIntrinsicsStream_->Intrinsics.Set(colorIntrinsics);
		cv::Mat test=colorIntrinsics.get_extrinsicMatrices();
    }

    return State::DEVICE_OPEN == state;
}

bool Freenect2Cam::Disconnect() {
    if (state != State::DEVICE_OPEN) {
        std::cerr << DEBUG_IDENTIFIER
             << "please make sure a Freenect2 device is connected before trying to disconnect"
             << std::endl;
        return false;
    }
    device->stop();
    device->close();
    delete device;
    device = nullptr;

    state = State::DEVICE_CLOSED;
    return true;
}

bool Freenect2Cam::onNewFrame(libfreenect2::Frame::Type type, libfreenect2::Frame *frame) {
    switch (type) {
        case libfreenect2::Frame::Color: {
            auto colorMat = matFromFrame(frame, colorStream_->m_cvPixelType()).clone();

            if (unmirror_) {
                cv::flip(colorMat, colorMat, 1);
            }

            colorSource_ << colorMat;


            // save for color depth mapping
            {
                std::unique_lock<std::mutex> lock(m_newColorAvailableMutex, std::defer_lock);
                if (!lock.try_lock()) {
                    std::cout << "skipping color" << std::endl;
                } else {
                    m_lastColor = colorMat;
                }
            }
            break;
        }
        case libfreenect2::Frame::Ir: {
            auto irMat = matFromFrame(frame, irStream_->m_cvPixelType()).clone();

            if (unmirror_) {
                cv::flip(irMat, irMat, 1);
            }

            ir_Source << irMat.clone();
            break;
        }
		case libfreenect2::Frame::Depth: {
			cv::Mat depthMat = matFromFrame(frame, depthStream_->m_cvPixelType()).clone();

			if (unmirror_) {
				cv::flip(depthMat, depthMat, 1);
			}

			depthSource_ << depthMat;
            break;
        }
    }
    return false; // false -> frame memory will be freed
}

Freenect2Cam::~Freenect2Cam() {
    Disconnect();
}

void Freenect2Cam::SetUnmirror(const bool unmirror) {
    unmirror_ = unmirror;
}

const bool Freenect2Cam::GetUnmirror() {
    return unmirror_;
}

} // namespace dyncam