/// @file
/// @author     Mayank Sharma
/// @version    ...
///
/// @section LICENSE
///
/// ...
///
/// @section DESCRIPTION
///
/// This file declares
/// <ul>
/// <li> Freenect2Cam : Class representing a Freenect2 camera.</li>
/// </ul>
/// It exposes the data streams from the camera through Reactive Event Streams
/// described by CppReact.

#pragma once

#include "../../ReactDomain.h"

#include <libfreenect2/libfreenect2.hpp>
#include <libfreenect2/frame_listener.hpp>
#include <libfreenect2/registration.h>
#include <libfreenect2/packet_pipeline.h>
#include <libfreenect2/logger.h>

#include <opencv2/opencv.hpp>
#include <vector>
#include <mutex>

#include "../BaseCam.h"
#include "src/Core/ImageStream.h"
#include "src/Core/IntrinsicsStream.h"

/// exposes freenect2 support for dyncam
namespace dyncam {

/// Class representing a Freenect2 Camera. It wraps around freenect2::Freenect2 camera to make it reactive, multithreaded.
/// It also provides better control over when to connect, disconnect and what type to data to stream
/// Usage :
/// \code
///    void freenectLauncher(){
///        freenect2cam::Freenect2CamManager freenect2CamManager; // Freenect2CamManager is needed to create Freenect2Cam objects for connected kinect devices
///        std::vector<freenect2cam::Freenect2Cam*> freenectVector = freenect2CamManager.FindDevices(); // the call to FindDevices creates Freenect2Cam objects for connected devices
///        // loop to traverse through the connected devices
///         for(int i=0;i<freenectVector.size();i++){
///            freenect2cam::Freenect2Cam* freenect2Cam = freenectVector.at(i);
///            std::cout<<freenect2Cam->serial<<std::endl;
///            // Set up your listeners!! i.e. Register reactive observers or listeners that receive the cv::Mat created for Depth, IR and Color Frames streamed for the current Freenect2Cam
///            Observe(freenect2Cam->depthEventStream,[] (cv::Mat depthMat){
///                cv::imshow("depthWin", depthMat);
///                cv::waitKey(1);
///            });
///            Observe(freenect2Cam->colorEventStream,[] (cv::Mat colorMat){
///                cv::imshow("rgbWin", colorMat);
///                cv::waitKey(1);
///            });
///            Observe(freenect2Cam->irEventStream,[] (cv::Mat irMat){
///                cv::imshow("irWin",irMat);
///                cv::waitKey(1);
///            });
///             // Set up What you wanna listen to i.e. ConfigureStream will set StreamMode and determine which of the observers set in the lines of code above will get triggered
///            freenect2Cam->ConfigureStream(freenect2cam::StreamMode::DEPTH_ONLY);
///        }
///    }
/// \endcode
class Freenect2Cam : public BaseCam, libfreenect2::FrameListener {
public:

    virtual std::set<StreamTypes> getStreamCapabilities() override;
    virtual bool Open() override;
    virtual void Close() override;

    /// @param serial_ no of the libfreenect2::Freenect2 device this class wraps around
    /// @param pipeline libfreenect2::PacketPipeline to use when connecting to libfreenect2::Freenect2
    explicit Freenect2Cam(std::string serial_ = "",
                          libfreenect2::PacketPipeline *pipeline = new libfreenect2::OpenGLPacketPipeline());

    ~Freenect2Cam() override;

    void SetUnmirror(const bool unmirror);
    const bool GetUnmirror();

protected:

    bool onNewFrame(libfreenect2::Frame::Type type, libfreenect2::Frame *frame) override;

    // Object used to manage communications with the Kinect
    libfreenect2::Freenect2 freenect2_;

    // TODO: make uniqe ptr with custom deleter
    // Device used to communicate with Kinect
    libfreenect2::Freenect2Device *device = nullptr;

    std::unique_ptr<libfreenect2::Registration> registration;

    cv::Mat m_lastColor;
    std::mutex m_newColorAvailableMutex;

    // TODO: make shared ptr with custom deleter
    // CPU/GPU pipeline
    libfreenect2::PacketPipeline *pipeline_ = nullptr;

    std::shared_ptr<ImageStream> colorStream_;
    std::shared_ptr<ImageStream> irStream_;
    std::shared_ptr<ImageStream> depthStream_;
    std::shared_ptr<ImageStream> registeredColorStream_;
    std::shared_ptr<ImageStream> undistortedDepthStream_;
    std::shared_ptr<IntrinsicsStream> irIntrinsicsStream_;
    std::shared_ptr<IntrinsicsStream> colorIntrinsicsStream_;

    // kinect data is mirrored by default, unmirror for further processing?
    bool unmirror_ = true;

    // Serial no for this Freenect2Cam
    std::string serial_ = "";

public:
    EventSourceT<cv::Mat> colorSource_ = MakeEventSource<D, cv::Mat>();
    EventSourceT<cv::Mat> ir_Source    = MakeEventSource<D, cv::Mat>();
    EventSourceT<cv::Mat> depthSource_ = MakeEventSource<D, cv::Mat>();
    EventSourceT<cv::Mat> registeredColorSource_ = MakeEventSource<D, cv::Mat>();
    EventSourceT<cv::Mat> undistorteddepthSource_ = MakeEventSource<D, cv::Mat>();

    // joined color and depth events
    EventsT<std::pair<cv::Mat, cv::Mat>> registeredMaps_;

    // open and connects to libfreenect2::Freenect2Device
    bool Connect();
    bool Disconnect(); ///<stop streaming and close libfreenect2::Freenect2Device
};


} // namespace dyncam

