//
// Created by ascadian on 9/18/18.
//

#pragma once

#include <boost/asio.hpp>
#include <boost/tuple/tuple.hpp>
#include "src/Core/Serialization.h"

namespace dyncam {

class Connection {
protected:
    // The underlying socket.
    boost::asio::ip::tcp::socket socket_;

    size_t message_size_;
    boost::asio::mutable_buffer headerBuffer_= boost::asio::buffer(&message_size_, sizeof(message_size_));
    boost::asio::streambuf bodyBuffer_;

    typedef std::function<void(const boost::system::error_code &e)> Callback;
    Callback callback_;

public:
    explicit Connection(boost::asio::ip::tcp::socket socket);
    explicit Connection(boost::asio::io_service &service);

    /// Get the underlying socket. Used for making a connection or for accepting
    /// an incoming connection.
    boost::asio::ip::tcp::socket& socket();

    // Asynchronously write a data structure to the socket.
    template<typename T>
    void async_write(const T &t, Callback &&callback) {
        callback_ = callback;

        // Serialize the data first so we know how large it is.
        std::ostream os(&bodyBuffer_);
        cereal::PortableBinaryOutputArchive archive(os);
        archive(t);

        size_t bodySize = bodyBuffer_.size();
        //std::cout << "Body size: " << bodySize << std::endl;


        // Write the serialized data to the socket. We use "gather-write" to send
        // both the header and the data in a single write operation.
        std::vector<boost::asio::const_buffer> buffers;
        buffers.push_back(boost::asio::buffer(&bodySize, sizeof(bodySize)));
        buffers.push_back(bodyBuffer_.data());
        boost::asio::async_write(socket_, buffers,
                                 [this, bodySize](const boost::system::error_code &e, std::size_t bytes_transferred) {
                                     // remove data we saved before. Use actual written size as we
                                     // may have an error during sending
                                     bodyBuffer_.consume(bodySize);
                                     callback_(e);
                                 });
    }

    /// Asynchronously read a data structure from the socket.
    template<typename T>
    void async_read(T &t, Callback &&callback) {
        callback_ = callback;
        boost::asio::async_read(socket_, headerBuffer_,
            [this, &t] (boost::system::error_code ec, std::size_t length) {
                handle_read_header(ec, t);
            });
    }

    /// Handle a completed read of a message header.
    template<typename T>
    void handle_read_header(const boost::system::error_code &e, T &t) {
        if (e) {
            callback_(e);
        } else {
            // Start an asynchronous call to receive the data.
            //std::cout << "reading body: " << message_size_ << std::endl;
            //std::cout << "buffer size: " << bodyBuffer_.size() << std::endl;
            boost::asio::async_read(socket_, bodyBuffer_, boost::asio::transfer_exactly(message_size_),
                [this, &t](boost::system::error_code ec, std::size_t length)
                {
                    //std::cout << "Body read: " << length << std::endl;
                    //std::cout << "bbuffer size: " << bodyBuffer_.size() << std::endl;
                    handle_read_body(ec, t);
                });
        }
    }

    /// Handle the body read.
    template<typename T>
    void handle_read_body(const boost::system::error_code &e, T &t) {
        if (e) {
            callback_(e);
        } else {
            // Extract the data structure from the data just received.
            std::istream is(&bodyBuffer_);
            cereal::PortableBinaryInputArchive inputArchive(is);
            inputArchive(t);
            //std::cout << "bbuffer size read body: " << bodyBuffer_.size() << std::endl;
            // Inform caller that data has been received ok.
            callback_(e);
        }
    }
};

} // namespace dyncam