//
// Created by ascadian on 12/1/17.
//

#include <iostream>
#include <src/Core/ImageStream.h>
#include <src/Core/IntrinsicsStream.h>
#include "ServerSession.h"
#include "Connection.h"
#include "StreamServer.h"
#include "src/ReactDomain.h"
#include "RequestStreams.h"
#include "CompressedStream.h"

namespace dyncam {

ServerSession::ServerSession(boost::asio::io_service &io_service, StreamServer *server) :
    m_connection(std::make_unique<Connection>(io_service)),
    m_server(server) {
    std::cout << "Serv: New Session" << std::endl;
}

ServerSession::~ServerSession() {
    std::cout << "Serv: Session closed" << std::endl;
}

boost::asio::ip::tcp::socket &ServerSession::socket() {
    return m_connection->socket();
}

void ServerSession::start() {
    readRequest();
}

void ServerSession::readRequest() {
    std::cout << "Serv: reading RequestStreams" << std::endl;
    std::shared_ptr<RequestStreams> req = std::make_shared<RequestStreams>();
    auto self(shared_from_this());
    m_connection->async_read(*req,
        [this, req=req, self](const boost::system::error_code &e) {
            if (e) {
                std::cerr << "ServerSession: " << e.message() << std::endl;
            } else {
                std::cout << "Server received: " << std::endl << *req << std::endl;

                if (req->isCollectionRequest()) {
                    sendStreamColl();
                } else {
                    startStreaming(req);
                }
            }
        });
}

void ServerSession::sendStreamColl() {
    std::cout << "Serv: sending StreamColl" << std::endl;
    auto self(shared_from_this());
    m_connection->async_write(*m_server->m_streams,
        [self] (const boost::system::error_code &e) {
          std::cout << "Serv: StreamColl send: " << std::endl;
        });
}

void ServerSession::startStreaming(std::shared_ptr<RequestStreams> request) {
    std::cout << "Serv: start Streaming: " << request << std::endl;
    // FIXME: here Image stream is hard coded!
    auto type = request->getType();

    switch (type) {
        case StreamTypes::UNKNOWN:
            break;
        case StreamTypes::COLOR:
        case StreamTypes::IR:
        case StreamTypes::DEPTH:
        case StreamTypes::UNDISTORTED_DEPTH:
        case StreamTypes::REGISTERED_COLOR:
            imageStream(request);
            break;
        case StreamTypes::EXTRINSICS:
            break;
        case StreamTypes::INTRINSICS:
            intrinsicsStream(request);
            break;
        case StreamTypes::COMPRESSED:
            compressedStream(request);
            break;
    }


}

void ServerSession::imageStream(std::shared_ptr<RequestStreams> request) {
    auto stream = m_server->m_streams->GetStreamById<ImageStream>(request->getStreamID());
    std::cout << "Serv: serving ImageStream" << std::endl;
    auto self(shared_from_this());
    m_activeStreams.emplace_back(Observe(stream->events,
        [this, self](const cv::Mat m) {
         m_connection->async_write(m,
            [self, m] (const boost::system::error_code &e) {
               std::cout << "Serv: img sent " << std::endl;
            });
        }));
}

void ServerSession::intrinsicsStream(std::shared_ptr<RequestStreams> request) {
    auto stream = m_server->m_streams->GetStreamById<IntrinsicsStream>(request->getStreamID());
    std::cout << "Serv: serving IntrinsicsStream" << std::endl;
    auto self(shared_from_this());
    m_activeStreams.emplace_back(Observe(stream->Intrinsics,
        [this, self](const CamIntrinsics m) {
         m_connection->async_write(m,
            [self, m] (const boost::system::error_code &e) {
               std::cout << "Serv: intrinsics sent " << std::endl;
            });
        }));
}

void ServerSession::compressedStream(std::shared_ptr<RequestStreams> request) {
    std::cout << "Serv: serving CompressedStream" << std::endl;
    auto stream = m_server->m_streams->GetStreamById<ImageStream>(request->getStreamID());
    auto compressedStream = std::make_shared<CompressedStream>(stream, request->GetCompression());
    auto self(shared_from_this());
    m_activeStreams.emplace_back(Observe(compressedStream->m_compression->compress(compressedStream->m_innerStream->events),
        [this, self, compressedStream](const cv::Mat m) {
         m_connection->async_write(m,
             [self, m] (const boost::system::error_code &e) {
                 std::cout << "Serv: compressed sent " << std::endl;
             });
        }));
}

} // namespace dyncam {