//
// Created by ascadian on 12/16/17.
//

#include "Lz4Compression.h"
#include <opencv2/opencv.hpp>
#include "lz4.h"

namespace dyncam {

Lz4Compression::Lz4Compression(const cv::Size m_outSize, const int m_outType) :
                                                                             outSize(m_outSize),
                                                                             outType(m_outType) {}


EventsT<cv::Mat> Lz4Compression::compress(const EventsT<cv::Mat> &inStream) {
    auto trans = Transform(inStream, [] (const cv::Mat &image) {

        const int src_size = image.elemSize() * image.rows * image.cols;
        // LZ4 provides a function that will tell you the maximum size of compressed output based on input data via LZ4_compressBound().
        const int max_dst_size = LZ4_compressBound(src_size);
        // We will use that size for our destination boundary when allocating space.
        std::vector<char> buf;
        buf.resize(max_dst_size / sizeof(char));
        // That's all the information and preparation LZ4 needs to compress *src into *compressed_data.
        // Invoke LZ4_compress_default now with our size values and pointers to our memory locations.
        // Save the return value for error checking.
        const int acceleration = 50; // 1 default >1 faster
        const int compressed_data_size = LZ4_compress_fast(
                reinterpret_cast<const char *>(image.data), buf.data(), src_size, max_dst_size, acceleration);
        // Check return_value to determine what happened.
        if (compressed_data_size < 0)
            throw std::runtime_error("A negative result from LZ4_compress_default indicates a failure trying to compress the data.  See exit code (echo $?) for value returned.");
        if (compressed_data_size == 0)
            throw std::runtime_error("A result of 0 means compression worked, but was stopped because the destination buffer couldn't hold all the information.");

        // set the exact size so everything gets copied into the mat
        buf.resize(compressed_data_size / sizeof(char));
        // copy data into mat
		cv::Mat result(buf, true);
        return result;
    });
    D::SetNodeName(*trans.ptr_, "Lz4Compression::compress");
    return trans;
}

EventsT<cv::Mat> Lz4Compression::deCompress(const EventsT<cv::Mat> &inStream) {
    auto trans = Transform(inStream, [&] (const cv::Mat &image) {
        cv::Mat out(outSize, outType);
        const int compressedSize = image.elemSize() * image.rows * image.cols;
        const int outBufferSize = out.elemSize() * out.rows * out.cols; 
		
		const int decompressed_size =  LZ4_decompress_safe(
                reinterpret_cast<const char *>(image.data), reinterpret_cast<char *>(out.data),
                compressedSize, outBufferSize);
		if (decompressed_size < 0)
            throw std::runtime_error("A negative result from LZ4_decompress_safe indicates a failure trying to decompress the data.  See exit code (echo $?) for value returned.");
        if (decompressed_size == 0)
            throw std::runtime_error("I'm not sure this function can ever return 0.  Documentation in lz4.h doesn't indicate so.");

        return out;
    });
    D::SetNodeName(*trans.ptr_, "Lz4Compression::deCompress");
    return trans;
}
std::string Lz4Compression::ToString() {
    return "Lz4Compression";
}

} // namespace dyncam