//
// Created by ascadian on 12/16/17.
//

#include "Lz4QuantCompression.h"
#include <opencv2/opencv.hpp>
#include "lz4.h"

namespace dyncam {

Lz4QuantCompression::Lz4QuantCompression(const cv::Size m_outSize, const int m_outType) :
                                                                             outSize(m_outSize),
                                                                             outType(m_outType) {}


EventsT<cv::Mat> Lz4QuantCompression::compress(const EventsT<cv::Mat> &inStream) {
    auto trans = Transform(inStream, [] (const cv::Mat &image) {

		//auto start = std::chrono::system_clock::now();

		//quantized compression
		cv::Mat fullrange = image*8.191875; //65535 (max value 16 bit int) / 8000 (max dist in cm of kinect v2) = 8.191875
		
		// replace NaN by 0
		//fullrange.setTo(0, fullrange != fullrange); //TODO when needed?
		
		cv::Mat target;
		switch (image.channels()) {
		case 1: {
			fullrange.convertTo(target, CV_16UC1);
			break;
		}
		case 3: {
			fullrange.convertTo(target, CV_16UC3);
			break;
		}
		case 4: {
			fullrange.convertTo(target, CV_16UC4);
			break;
		}
		default:
			std::cerr << "Mat not supported by QuantizedCompression: channels" << image.channels() << std::endl;
			break;
		}

		//LZ4 compression
        const int src_size = target.elemSize() * target.rows * target.cols;
        // LZ4 provides a function that will tell you the maximum size of compressed output based on input data via LZ4_compressBound().
        const int max_dst_size = LZ4_compressBound(src_size);
        // We will use that size for our destination boundary when allocating space.
        std::vector<char> buf;
        buf.resize(max_dst_size / sizeof(char));
		// That's all the information and preparation LZ4 needs to compress *src into *compressed_data.
        // Invoke LZ4_compress_default now with our size values and pointers to our memory locations.
        // Save the return value for error checking.
        const int acceleration = 50; // 1 default >1 faster
        const int compressed_data_size = LZ4_compress_fast(
                reinterpret_cast<const char *>(target.data), buf.data(), src_size, max_dst_size, acceleration);
        
		// Check return_value to determine what happened.
        if (compressed_data_size < 0)
            throw std::runtime_error("A negative result from LZ4_compress_default indicates a failure trying to compress the data.  See exit code (echo $?) for value returned.");
        if (compressed_data_size == 0)
            throw std::runtime_error("A result of 0 means compression worked, but was stopped because the destination buffer couldn't hold all the information.");
		
		
		//std::cout << "bytes: " << compressed_data_size << std::endl;
        
		// set the exact size so everything gets copied into the mat
        buf.resize(compressed_data_size / sizeof(char));
        // copy data into mat
		cv::Mat result(buf, true);

		/*auto end = std::chrono::system_clock::now();
		auto diff = end - start;
		std::cout << "duration comp: " << std::chrono::duration_cast<std::chrono::microseconds>(diff).count() << std::endl;*/

        return result;
    });
    D::SetNodeName(*trans.ptr_, "Lz4QuantCompression::compress");
    return trans;
}

EventsT<cv::Mat> Lz4QuantCompression::deCompress(const EventsT<cv::Mat> &inStream) {
    auto trans = Transform(inStream, [&] (const cv::Mat &image) {

		//auto start = std::chrono::system_clock::now();

		//LZ4 decompression
        cv::Mat out(outSize, outType);
        const int compressedSize = image.elemSize() * image.rows * image.cols;
        const int outBufferSize = out.elemSize() * out.rows * out.cols; 
		
		const int decompressed_size =  LZ4_decompress_safe(
                reinterpret_cast<const char *>(image.data), reinterpret_cast<char *>(out.data),
                compressedSize, outBufferSize);
		
		if (decompressed_size < 0)
            throw std::runtime_error("A negative result from LZ4_decompress_safe indicates a failure trying to decompress the data.  See exit code (echo $?) for value returned.");
        if (decompressed_size == 0)
            throw std::runtime_error("I'm not sure this function can ever return 0.  Documentation in lz4.h doesn't indicate so.");

		//quantized decompression
		cv::Mat target;
		switch (out.channels()) {
		case 1: {
			out.convertTo(target, CV_32FC1);
			break;
		}
		case 3: {
			out.convertTo(target, CV_32FC3);
			break;
		}
		case 4: {
			out.convertTo(target, CV_32FC4);
			break;
		}
		default:
			std::cerr << "Mat not supported by QuantizedCompression: channels" << out.channels() << std::endl;
			break;
		}

		cv::Mat originValues = target / 8.191875;

		/*auto end = std::chrono::system_clock::now();
		auto diff = end - start;
		std::cout << "duration decomp: " << std::chrono::duration_cast<std::chrono::microseconds>(diff).count() << std::endl;*/

        return originValues;
    });
    D::SetNodeName(*trans.ptr_, "Lz4QuantCompression::deCompress");
    return trans;
}
std::string Lz4QuantCompression::ToString() {
    return "Lz4QuantCompression";
}

} // namespace dyncam