//
// Created by ascadian on 8/8/17.
//

#include "CloudFromDepth.h"
#include "src/Core/ImageStream.h"
#include "src/Core/IntrinsicsStream.h"

namespace dyncam {

	CloudFromDepth::CloudFromDepth(StreamCollection &streams) {

		// use undistorted depth if available
		auto depth = streams.GetStreamFirst<ImageStream>(StreamTypes::UNDISTORTED_DEPTH);
		if (depth) {
			undistort_depth_ = false;
		}
		else {
			// fallback for no undistorted stream. Use raw stream and undistort on the fly
			depth = streams.GetStreamFirst<ImageStream>(StreamTypes::DEPTH);
			undistort_depth_ = true;
		}
		assert(nullptr != depth);

		// the intrinsics are in the ir stream
		const auto rawIr = streams.GetStreamFirst<ImageStream>(StreamTypes::IR);
		assert(nullptr != rawIr);
		auto intrinsicsParentId = rawIr->GetId();
		auto intrinsics = streams.GetStreamFirst<IntrinsicsStream>(StreamTypes::INTRINSICS, intrinsicsParentId);
		assert(nullptr != intrinsics);

		PointPosEvents = CloudFromDepth::transformDepth(depth->events, intrinsics->Intrinsics, undistort_depth_);
		D::SetNodeName(*PointPosEvents.ptr_, "CloudFromDepth::PointPosEvents");

		//use registered color if available
		auto regColor = streams.GetStreamFirst<ImageStream>(StreamTypes::REGISTERED_COLOR); 
		if (regColor) {
			reg_color = false;
		}
		else {
			reg_color = true;
			regColor= streams.GetStreamFirst<ImageStream>(StreamTypes::COLOR);
		}
		assert(nullptr != regColor);

		// the intrinsics are in the color stream
		const auto color = streams.GetStreamFirst<ImageStream>(StreamTypes::COLOR);
		assert(nullptr != color);
		auto intrinsicsParentIdC = color->GetId();
		auto colorIntrinsics = streams.GetStreamFirst<IntrinsicsStream>(StreamTypes::INTRINSICS, intrinsicsParentIdC);
		assert(nullptr != colorIntrinsics);

		PointColorsEvents = transformColor(regColor->events, colorIntrinsics->Intrinsics, reg_color);
		D::SetNodeName(*PointColorsEvents.ptr_, "CloudFromDepth::PointColorsEvents");
	}



inline void
CloudFromDepth::rs_deproject_pixel_to_point(const CamIntrinsics &intrin, const bool undistort,
                                            const int row, const int col, const float depth_value,
                                            float *res) {
    auto depth = depth_value / 1000.0f; // scale to 1m
    const float bad_point = std::numeric_limits<float>::quiet_NaN();
    if (isnan(depth) || depth <= 0.001 ) { 
        //depth_value value is not valid
        res[0] = bad_point;
        res[1] = bad_point;
        res[2] = bad_point;
    } else {
        // 3d camera projection
        float x = (row + 0.5f - intrin.get_cx()) / intrin.get_fx();
        float y = (col + 0.5f - intrin.get_cy()) / intrin.get_fy();

        if (undistort) {
            // undistort
            float r2 = x * x + y * y;
            float f = 1 + intrin.get_k1() * r2 + intrin.get_k2() * r2 * r2 + intrin.get_k3() * r2 * r2 * r2;
            float ux = x * f + 2 * intrin.get_p1() * x * y + intrin.get_p2() * (r2 + 2 * x * x);
            float uy = y * f + 2 * intrin.get_p2() * x * y + intrin.get_p1() * (r2 + 2 * y * y);
            x = ux;
            y = uy;
        }

        res[0] = depth * x;
        res[1] = depth * y;
        res[2] = depth;
    }
}

EventsT<cv::Mat>
CloudFromDepth::transformColor(EventsT<cv::Mat> &color, VarSignalT<CamIntrinsics> &intrinsics, const bool regColor) {
    auto transf = Transform(color, [regColor, &intrinsics](const cv::Mat &image) {
        // remove alpha and reorder channels
        cv::Mat out;
        if (image.channels() == 4) {
            cv::cvtColor(image, out, cv::COLOR_BGRA2RGB);
        }
        else {
            cv::cvtColor(image, out, cv::COLOR_BGR2RGB);
        }
        assert(out.channels() == 3);

        if (regColor)
        {
            //TODO register color manually using intrinsics
        }

        // returns CV_8UC3
        return out;
    });
    D::SetNodeName(*transf.ptr_, "CloudFromDepth::transformColor");
    return transf;
}

EventsT<cv::Mat>
CloudFromDepth::transformDepth(EventsT<cv::Mat> &depth, VarSignalT<CamIntrinsics> &intrinsics, const bool undistort) {
    auto transf = Transform(depth, With(intrinsics),
                     [undistort = undistort](const cv::Mat &image, const CamIntrinsics &intr) {
        const auto cols = image.cols;
        const auto rows = image.rows;
        const auto points = rows * cols;
        cv::Mat res(points, 3, CV_32F);
        // process each pixel in parallel
        //typedef cv::TypeDepth<CV_32FC1>::value_type data_type;
        using data_type = float;

        image.forEach<data_type>(
                [cols, &intr, &res, undistort](const data_type& pixel, const int position[]) -> void {
                    // rows cols are change. opencv (.at(y,x)) or other reason?
                    const int y = position[0];
                    const int x = position[1];
                    // linearize point cloud
                    const int i = (x + y * cols);
                    auto addr = &res.at<float>(i, 0);
                    // last parameter is the adress where the result triple will be written to
                    CloudFromDepth::rs_deproject_pixel_to_point(intr, undistort, x, y, pixel,
                                                                addr);
                }
        );

//        for (int r = 0; r < image.rows; ++r) {
//            for (int c = 0; c < image.cols; ++c) {
//                const int y = r;
//                const int x = c;
//                // linearize point cloud
//                const int i = (x + y * cols);
//                auto addr = &res.at<float>(i, 0);
//                // last parameter is the adress where the result triple will be written to
//                data_type pixel = image.at<data_type>(y, x);
//                CloudFromDepth::rs_deproject_pixel_to_point(intr, x, y, pixel, addr);
//            }
//        }

		return res;
    });
    D::SetNodeName(*transf.ptr_, "CloudFromDepth::transformDepth");
    return transf;
}

} // namespace dyncam